#ifndef COMMON
#define COMMON

/**
 * This model of ALICE Data Point Service (ADAPOS) has four parameters, the
 * number of Engines, the number of Terminals, the number of services, and the
 * number of generated events. The LTL properties of this model don't scale with
 * the number of Engines and Terminals. Use ADAPOS theory generator for
 * generating scaled theories.
 */
#ifndef ENGINE_COUNT
#define ENGINE_COUNT   2
#endif
#ifndef TERMINAL_COUNT
#define TERMINAL_COUNT 2
#endif
#ifndef SERVICE_COUNT
#define SERVICE_COUNT  3
#endif
#ifndef EVENT_COUNT
/* (EVENT_COUNT / SERVICE_COUNT) * SERVICE_COUNT + SERVICE_COUNT - 1
   must be at most 125. */
#define EVENT_COUNT    10
#endif
#define PROCESS_COUNT  (ENGINE_COUNT + TERMINAL_COUNT)

#define MSG_DISCONNECT 126
#define MSG_NONE       127

#define FBI_BIT        128
#define PAYLOAD_MASK   127

#ifdef NDEBUG
#define ASSERT(e) skip
#else
#define ASSERT(e) assert(e)
#endif

/**
 * Global naming convention:
 *   - i runs over Engines
 *   - j runs over Terminals
 *   - x runs over either Engines or Terminals
 *   - k runs over x's FBI
 *   - z runs over events (in Load Generator)
 *   - m denotes messages
 *   - v denotes values
 */

/**
 * Increments the value n by one, modulo c.
 */
#define INCREMENT(n, c)                                                        \
  if                                                                           \
    :: n + 1 >= c ->                                                           \
       n = 0                                                                   \
    :: else ->                                                                 \
       n = n + 1                                                               \
  fi

/**
 * Converts a Terminal index number to ADAPOS process index number for accessing
 * the global protocol state array.
 */
#define OFFSET(j)                                                              \
  (ENGINE_COUNT + j)

/**
 * Translates the 2D coordinates of FBI into a proper index value.
 */
#define TRANSLATE_2D(x, k)                                                     \
  (x * SERVICE_COUNT) + k

/**
 * Translates the 3D coordinates of the message array into a proper index value.
 */
#define TRANSLATE_3D(i, j, k)                                                  \
  (i * TERMINAL_COUNT + j) * SERVICE_COUNT + k

/**
 * There are two possible protocol states for an ADAPOS application in this
 * model: SEND_SINGLE and SEND_FBI. During SEND_SINGLE, single updates to data
 * points will be transmitted and during SEND_FBI a whole FBI will be
 * transmitted.
 */
mtype = { SEND_SINGLE, SEND_FBI };

/**
 * The global protocol state array. Indexed the same way as the ADAPRO state
 * array.
 */
mtype protocolStates [ PROCESS_COUNT ];

/**
 * True iff LG has stopped producing updates.
 */
bool end = false;

/**
 * An array representing states of data points published to ADAPOS through DIM.
 */
byte inputDataPoints [ SERVICE_COUNT ];

/**
 * Represents the current and the previous state of a datapoint. A boolean is
 * used for indicating whether or not the state has been obtained from an FBI
 * transmission.
 */
typedef DPstate
{
  byte current;
  byte previous;
  bool fresh;
}

/**
 * An array representing the states of FBIs of Terminals. This array is indexed
 * in a two-dimensional fashion.
 */
DPstate terminalFBIs [ TERMINAL_COUNT * SERVICE_COUNT ];

/**
 * An array representing the states of the 0MQ connections between Engines and
 * Terminals. It is indexed in a three-dimensional fashion.
 */
byte messages        [ ENGINE_COUNT * TERMINAL_COUNT * SERVICE_COUNT ];

#endif /* COMMON */