#ifndef ENGINE
#define ENGINE

#include "./Common.pml"

/**
 * Simulates sending the given message with the given message from Engine i to
 * Terminal j. The index k ranges over the FBI and m is the message content.
 * The word "write" is semantically more accurate than "send", because the
 * underlying mechanism of message transport in this (variant of the) model is
 * based on arrays rather than channels.
 *
 * According to 0MQ documentation: "When a PUB socket enters the mute state due
 * to having reached the high water mark for a subscriber, then any messages
 * that would be sent to the subscriber in question shall instead be dropped
 * until the mute state ends." This behaviour is simulated by discarding the
 * message m if the messages array already contains a message other than
 * MSG_NONE.
 */
#define WRITE_MESSAGE(i, j, k, m)                                              \
  d_step                                                                       \
  {                                                                            \
    if                                                                         \
      :: messages[ TRANSLATE_3D(i, j, k) ] == MSG_NONE ->                      \
         messages[ TRANSLATE_3D(i, j, k) ] = m                                 \
      :: else ->                                                               \
         printf("A message from Engine %d to Terminal %d was lost.\n", i, j)   \
    fi                                                                         \
  }

/**
 * Multicasts the given message m at FBI index k from Engine i to all Terminals.
 * Connections may die non-deterministically, which is modelled by sending
 * disconnect messages.
 */
inline sendMulticast(i, j, k, m)
{
  for (j : 0 .. TERMINAL_COUNT - 1)
  {
    if
      :: WRITE_MESSAGE(i, j, k, m)
      :: messages[ TRANSLATE_3D(i, j, k) ] = MSG_DISCONNECT
    fi
  }
}

/**
 * Processes a single update by polling all indices in the input array. The
 * update will be multicasted to Terminals. NB: Real ADAPOS Engine doesn't poll
 * DIM services. Rather, it executes a callback whenever an update arrives.
 */
inline singleUpdate(i, j, k)
{
  skip;
endWaitForLG: end == false;
  if
    :: inputDataPoints[k] > fbi[k] ->
       printf("Engine %d writes %d into its FBI at index %d.\n",
	      i,
	      inputDataPoints[k],
	      k
       );
       fbi[k] = inputDataPoints[k];
       sendMulticast(i, j, k, inputDataPoints[k])
    :: else ->
       skip
  fi
  INCREMENT(k, SERVICE_COUNT)
}

/**
 * Checks if there is capacity for sending an FBI to Terminal j. If there isn't,
 * then the value of k will be set to SERVICE_COUNT + 1.
 */
inline checkCapacity(i, j, k)
{
  k = 0
  do
    :: k < SERVICE_COUNT ->
       if
	 :: messages[ TRANSLATE_3D(i, j, k) ] != MSG_NONE ->
	    /* There is at least one message still in transit. This
                     cancels FBI transmission from Engine i to Terminal j. */
	    k = SERVICE_COUNT + 1
	 :: else ->
	    skip
       fi
       k = k + 1
    :: else ->
       break
  od
}

/**
 * Sends the FBI to all Terminals. Any given Terminal either receives the FBI
 * fully or not at all.
 */
inline sendFBI(i, j, k)
{
  printf("Engine(i: %d, j: %d, k: %d, s: %d): sendFBI\n",
	 i, j, k, protocolStates[i]);
  for (j : 0 .. TERMINAL_COUNT - 1)
  {
    if
      :: d_step
	 {
	   checkCapacity(i, j, k);
	   if
	     :: k == SERVICE_COUNT ->
		for (k : 0 .. SERVICE_COUNT - 1)
		{
		  messages[ TRANSLATE_3D(i, j, k) ] = (fbi[k] | FBI_BIT)
		}
		skip
	     :: else ->
		skip
	   fi
	   k = 0
	 }
      :: messages[ TRANSLATE_3D(i, j, 0) ] = MSG_DISCONNECT
    fi
  }
  skip;
  d_step
  {
    k = 0;
    protocolStates[i] = SEND_SINGLE
  }
}

/**
 * Simulates the logic of ADAPOS Engine during its ADAPRO state RUNNING.
 */
inline executeEngine(i, j, k)
{
  do
    :: atomic
       {
	 printf("Engine(i: %d, j: %d, k: %d, s: %d): executeEngine\n",
		i, j, k, protocolStates[i]);
	 if
	   :: protocolStates[i] == 0 -> /* Startup */
	      protocolStates[i] = SEND_SINGLE
	   :: protocolStates[i] == SEND_SINGLE ->
	      singleUpdate(i, j, k)
	   :: protocolStates[i] == SEND_FBI    ->
	      sendFBI(i, j, k)
	 fi
       }
  od
}

/**
 * The ADAPOS Engine process type. The argument i is the index of the Engine
 * instance and the argument k is an index used for iterating over services
 * and Terminals.
 */
proctype Engine(byte i)
{
  byte j;
  byte k;
  byte fbi [ SERVICE_COUNT ];
execEng:   executeEngine(i, j, k);
}

#endif /* ENGINE */