#ifndef TERMINAL
#define TERMINAL

#include "./Common.pml"

/**
 * Reads the element k from the FBI of Terminal j.
 */
#define READ_FBI(j, k)                                                         \
  terminalFBIs[ TRANSLATE_2D(j, k) ].current

/**
 * A version of READ_FBI that takes the payload mask.
 */
#define READ_FBI_PAYLOAD(j, k)                                                 \
  (READ_FBI(j, k) & PAYLOAD_MASK)

/**
 * This macro writes the value v to index k in the FBI of Terminal j.
 */
#define WRITE_FBI(j, k, v)                                                     \
  terminalFBIs[ TRANSLATE_2D(j, k) ] .previous =                               \
    terminalFBIs[ TRANSLATE_2D(j, k) ].current;                                \
  terminalFBIs[ TRANSLATE_2D(j, k) ].current = v

/**
 * Blocks until a message arrives from Engine i to Terminal j.
 */
#define WAIT_FOR_MESSAGE(i, j, k)                                              \
  messages[ TRANSLATE_3D(i, j, k) ] != MSG_NONE

/**
 * Reads the latest message transmitted from process i to j. The index k ranges
 * over the FBI.
 */
#define PEEK_MESSAGE(i, j, k)                                                  \
  messages[ TRANSLATE_3D(i, j, k) ]

/**
 * Reads the payload part of the latest message transmitted from Engine i to
 * Terminal j. The index k ranges over the FBI.
 */
#define PEEK_PAYLOAD(i, j, k)                                                  \
  (messages[ TRANSLATE_3D(i, j, k) ] & PAYLOAD_MASK)

/**
 * Copies the current message from Engine i to Terminal j into Terminal j's FBI
 * at index k.
 */
#define COPY_MESSAGE(i, j, k)                                                  \
  printf("Terminal %d writes %d into its FBI at index %d.\n",                  \
         j,                                                                    \
         PEEK_PAYLOAD(i, j, k),                                                \
         k);                                                                   \
  WRITE_FBI(j, k, PEEK_PAYLOAD(i, j, k))


/**
 * Clears a message identified by indices i, j, and k in from the message array.
 */
#define CLEAR_MESSAGE(i, j, k)                                                 \
  messages[ TRANSLATE_3D(i, j, k) ] = MSG_NONE

/**
 * Connect to an Engine. Never fails.
 */
inline connect(i, j, k)
{
  printf("Terminal(i: %d, j: %d, k: %d, s: %d): connect\n",
	 i, j, k, protocolStates[OFFSET(j)]);
  /* All old messages will be discarded: */
  for (k : 0 .. SERVICE_COUNT - 1)
  {
    CLEAR_MESSAGE(i, j, k)
  }
  /* Terminal changes its own state: */
  protocolStates[OFFSET(j)] = SEND_FBI;
  /* Through DIM, Terminal tells Engine to change its state: */
  protocolStates[i] = SEND_FBI;
  k = 0
  printf("Terminal %d connected to Engine %d.\n", j, i);
}

/**
 * Simulates the event when the connection between Engine i and Terminal j dies
 * for any reason. This model doesn't differentiate between different possible
 * causes (e.g. network or Engine failure) for a connection terminating.
 */
inline disconnect(i, j, k)
{
  printf("Terminal %d disconnected from Engine %d.\n", j, i);
  INCREMENT(i, ENGINE_COUNT);
}

/**
 * Disconnects from the current Engine and connects to the next Engine.
 */
inline reconnect(i, j, k)
{
  disconnect(i, j, k);
  connect(i, j, k)
}

/**
 * Checks if the FBI was transmitted successfully. If it wasn't then, k will be
 * set to SERVICE_COUNT + 1.
 */
inline checkFBI(i, j, k)
{
  k = 0;
  do
    :: k < SERVICE_COUNT ->
       if
	 :: (PEEK_MESSAGE(i, j, k) & FBI_BIT) == 0  ||
	    PEEK_PAYLOAD(i, j, k) == MSG_DISCONNECT ->
	    /* Terminal interprets a partial FBI as a connection failure. */
	    printf("The FBI Terminal %d received was partial or the connection timed out.\n", j);
	    k = SERVICE_COUNT + 1
	 :: else ->
	    skip
	 fi
       k = k + 1
    :: else ->
       break
  od
}

/**
 * Copies all messages into the Terminal's own FBI.
 */
inline copyFBI(i, j, k)
{
  for (k : 0 .. SERVICE_COUNT - 1)
  {
    COPY_MESSAGE(i, j, k);
    terminalFBIs[TRANSLATE_2D(j, k)].fresh = true;
    CLEAR_MESSAGE(i, j, k)
  }
}

/**
 * Processes an FBI.
 */
inline processFBI(i, j, k)
{
  printf("Terminal %d received FBI from Engine %d.\n", j, i);
  checkFBI(i, j, k);
  if
    :: k == SERVICE_COUNT ->
       copyFBI(i, j, k);
       protocolStates[OFFSET(j)] = SEND_SINGLE;
    :: else ->
       reconnect(i, j, k)
  fi
  k = 0
}

/**
 * Processes a single update.
 */
inline processSingle(i, j, k)
{
  if
    :: (PEEK_MESSAGE(i, j, k) & FBI_BIT) > 0 ->
       printf("The message was part of an unexpected FBI.\n");
       processFBI(i, j, k)
    :: else ->
       COPY_MESSAGE(i, j, k);
       terminalFBIs[TRANSLATE_2D(j, k)].fresh = false;
       CLEAR_MESSAGE(i, j, k);
       INCREMENT(k, SERVICE_COUNT)
  fi
}

/**
 * Processes a single update or an FBI.
 */
inline process(i, j, k)
{
  if
    :: protocolStates[OFFSET(j)] == SEND_FBI    ->
       processFBI(i, j, k)
    :: protocolStates[OFFSET(j)] == SEND_SINGLE ->
       processSingle(i, j, k)
  fi
}

/**
 * Simulates the logic of ADAPOS Terminal during its ADAPRO state RUNNING.
 */
inline executeTerminal(i, j, k)
{
  do
    :: atomic
       {
	 printf("Terminal(i: %d, j: %d, k: %d, s: %d): executeTerminal\n",
		i, j, k, protocolStates[OFFSET(j)]);
endWait: WAIT_FOR_MESSAGE(i, j, k);
	 d_step
	 {
	   printf("Terminal %d got message %d.\n", j, PEEK_MESSAGE(i, j, k));
	   if
	     :: PEEK_PAYLOAD(i, j, k) == MSG_DISCONNECT ->
		reconnect(i, j, k)
	     :: else ->
		process(i, j, k)
	   fi
	   skip
	 }
       }
  od
}

/**
 * The ADAPOS Terminal process type. The argument i is the index of the Engine
 * process that Terminal will connect to. Byte j is the index of the Terminal
 * process itself (starting from zero). Byte k is an index used for accessing
 * the Full Buffer Image (FBI).
 */
proctype Terminal(byte j)
{
  byte i;
  byte k;
prepTerm: d_step
  {
    connect(i, j, k)
  }
execTerm: executeTerminal(i, j, k)
}

#endif /* TERMINAL */